@@ -1,6 +1,6 @@
 # -*- coding: iso-8859-1 -*-
 # GNU Solfege - free ear training software
-# Copyright (C) 2001, 2002, 2003, 2004, 2007, 2008, 2011  Tom Cato Amundsen
+# Copyright (C) 2001, 2002, 2003, 2004, 2007, 2008, 2011, 2016  Tom Cato Amundsen
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 # 4.69
-from __future__ import absolute_import
+
 """
 prog             The test done before calling
  +statementlist
@@ -47,13 +47,6 @@ import sys
 import weakref
 
 from solfege import i18n
-
-# We had to do this to make running "test.py" and
-# "test.py test_lessonfile" succeed. Without this,
-# "test.py" will fail because _ will not strip away
-# the text before "|" in strings like "chord|m7"
-_ = i18n.translation.ugettext
-
 import solfege.parsetree as pt
 
 NAME = 'NAME'
@@ -73,7 +66,7 @@ NEW_re = re.compile("""(?:
                         ("(.*?)")| #string
                         (\w[\[\]\w-]*) #name
                 )""",
-                      re.VERBOSE|re.MULTILINE|re.DOTALL|re.UNICODE)
+                      re.VERBOSE | re.MULTILINE | re.DOTALL | re.UNICODE)
 
 LI_INTEGER = NEW_re.match("-3").lastindex
 LI_FLOAT = NEW_re.match("3.3").lastindex
@@ -102,18 +95,22 @@ TOKEN_STRING = 1
 TOKEN_IDX = 2
 TOKEN_LINENO = 3
 
-class istr(unicode):
+
+class istr(str):
+
     def __init__(self, s):
         self.cval = s
         self.m_added_language = None
+
     def __mod__(self, other):
         """
         Handle format strings in translated strings:
         _("%i. inversion") % 2
         """
-        i=istr(unicode(self) % other)
+        i = istr(str(self) % other)
         i.cval = self.cval % other
         return i
+
     def add_translation(self, lang, s):
         """
         Use this method to add translations that are included directly in
@@ -127,7 +124,7 @@ class istr(unicode):
             # The first language in the list is preferred.
             new_pos = i18n.langs().index(lang)
             if not self.m_added_language:
-                old_pos = sys.maxint
+                old_pos = sys.maxsize
             else:
                 old_pos = i18n.langs().index(self.m_added_language)
             if new_pos < old_pos:
@@ -136,12 +133,14 @@ class istr(unicode):
                 retval.cval = self.cval
                 return retval
         return self
+
     def new_translated(cval, translated):
         retval = istr(translated)
         retval.cval = cval
         return retval
     new_translated = staticmethod(new_translated)
 
+
 def dataparser_i18n_func(s):
     if s == "":
         s = 'error in your lesson file: empty string cannot be translated'
@@ -149,6 +148,7 @@ def dataparser_i18n_func(s):
     retval.cval = s
     return retval
 
+
 def dataparser_i18n__i_func(s):
     if s == "":
         s = 'error in your lesson file: empty string cannot be translated'
@@ -158,34 +158,42 @@ def dataparser_i18n__i_func(s):
 
 
 class Question(dict):
+
     def __getattr__(self, n):
         if n in self:
             return self[n]
         raise AttributeError()
+
     def __setattr__(self, name, value):
         self[name] = value
 
 
 class DataparserException(Exception):
+
     def __init__(self, message):
         Exception.__init__(self, message)
 
 
 class DataparserSyntaxError(DataparserException):
+
     def __init__(self, parser, bad_pos, expect):
         DataparserException.__init__(self, _('Syntax error in file "%(filename)s". %(expected)s') % {'filename': parser.m_filename, 'expected': expect})
         # This variable is only used by the module test code.
         self.m_token = parser._lexer.m_tokens[bad_pos]
         self.m_nonwrapped_text = parser._lexer.get_err_context(bad_pos)
 
+
 class AssignmentToReservedWordException(DataparserException):
+
     def __init__(self, parser, bad_pos, word):
         DataparserException.__init__(self, _("Assignment to the reserved word \"%(word)s\"") % {'word': word})
         # This variable is only used by the module test code.
         self.m_token = parser._lexer.m_tokens[bad_pos]
         self.m_nonwrapped_text = parser._lexer.get_err_context(bad_pos)
 
+
 class CanOnlyTranslateStringsException(DataparserException):
+
     def __init__(self, parser, bad_pos, variable):
         DataparserException.__init__(self, _("We can only translate strings using in-file translations (ex var[no]=...). See the variable \"%(variable)s\" in the file \"%(filename)s\"") % {'filename': parser.m_filename, 'variable': variable})
         # This variable is only used by the module test code.
@@ -194,6 +202,7 @@ class CanOnlyTranslateStringsException(DataparserException):
 
 
 class UnableToTokenizeException(DataparserException):
+
     def __init__(self, lexer, lineno, token, pos):
         """
         lineno is the zero indexed line number where the exception happened.
@@ -212,24 +221,54 @@ class UnableToTokenizeException(DataparserException):
         self.m_nonwrapped_text = lexer.get_tokenize_err_context()
 
 
+def read_encoding_marker_from_string(string):
+    """
+    According to http://www.python.org/dev/peps/pep-0263/
+    the encoding marker must be in the first two lines
+    Return None if no encoding marker found
+    """
+    r = re.compile("#.*?coding\s*[:=]\s*([\w_.-]+)")
+    c = 0
+    for line in string.split("\n"):
+        m = r.match(line)
+        if m:
+            return m.groups()[0]
+        c += 1
+        if c == 2:
+            break
+
+
+def read_encoding_marker_from_file(filename):
+    """
+    According to http://www.python.org/dev/peps/pep-0263/
+    the encoding marker must be in the first two lines
+    Return None if no encoding marker found
+    """
+    r = re.compile(b"#.*?coding\s*[:=]\s*([\w_.-]+)")
+    with open(filename, 'br') as f:
+        line = f.readline()
+        m = r.match(line)
+        if m:
+            return m.groups()[0].decode("ascii")
+        line = f.readline()
+        m = r.match(line)
+        if m:
+            return m.groups()[0].decode("ascii")
+        return
+
+
 class Lexer:
+
     def __init__(self, src, parser):
+        assert isinstance(src, str)
         if parser:
             self.m_parser = weakref.ref(parser)
         else:
             self.m_parser = parser
-        r = re.compile("#.*?coding\s*[:=]\s*([\w_.-]+)")
-        # according to http://www.python.org/dev/peps/pep-0263/
-        # the encoding marker must be in the first two lines
-        m = r.match("\n".join(src.split("\n")[0:2]))
-        if m:
-            src = unicode(src, m.groups()[0], errors="replace")
-        else:
-            src = unicode(src, "UTF-8", errors="replace")
         src = src.replace("\r", "\n")
         # Some editors (notepad on win32?) insert the BOM, so we have
         # to check for it and remove it since the lexer don't handle it.
-        src = src.lstrip(unicode(codecs.BOM_UTF8, "utf8"))
+        src = src.lstrip(str(codecs.BOM_UTF8, "utf8"))
         self.m_src = src
         self.pos = 0
         pos = 0
@@ -237,8 +276,8 @@ class Lexer:
         self.m_tokens = []
         while 1:
             try:
-                if src[pos] in u"\u202f\xa0 \n\t{}=%+,/().":
-                    if src[pos] in u'\u202f\xa0 \t':
+                if src[pos] in "\u202f\xa0 \n\t{}=%+,/().":
+                    if src[pos] in '\u202f\xa0 \t':
                         pos += 1
                         continue
                     if src[pos] == '\n':
@@ -263,29 +302,33 @@ class Lexer:
         self.m_tokens.append(("EOF", None, pos, lineno))
         self.m_tokens.append(("EOF", None, pos, lineno))
         self.m_tokens.append(("EOF", None, pos, lineno))
+
     def _err_context_worker(self, lexer_pos):
         ret = ""
         lineno = self.m_tokens[lexer_pos][TOKEN_LINENO]
         x = self.m_tokens[lexer_pos][TOKEN_IDX]
-        while x > 0 and self.m_src[x-1] != "\n":
+        while x > 0 and self.m_src[x - 1] != "\n":
             x -= 1
         linestart_idx = x
         erridx_in_line = self.m_tokens[lexer_pos][TOKEN_IDX] - linestart_idx
         if lineno > 1:
-            ret += "\n(line %i): %s" % (lineno-1, self.get_line(lineno-2))
+            ret += "\n(line %i): %s" % (lineno - 1, self.get_line(lineno - 2))
         if lineno > 0:
-            ret += "\n(line %i): %s" % (lineno, self.get_line(lineno-1))
+            ret += "\n(line %i): %s" % (lineno, self.get_line(lineno - 1))
         ret += "\n(line %i): %s" % (lineno + 1, self.get_line(lineno))
-        ret += "\n" + " " * (erridx_in_line + len("(line %i): " % (lineno+1))) + "^"
+        ret += "\n" + " " * (erridx_in_line + len("(line %i): " % (lineno + 1))) + "^"
         return ret.strip()
+
     def get_tokenize_err_context(self):
         """
         return a string with the last part of the file that we were able
         to tokenize. Used by UnableToTokenizeException
         """
-        return self._err_context_worker(len(self.m_tokens)-1)
+        return self._err_context_worker(len(self.m_tokens) - 1)
+
     def get_err_context(self, pos):
         return self._err_context_worker(pos)
+
     def new_get_err_context(self, pos1, pos2):
         # Line number of the last part of the error. We will display
         # two lines, the last line containing (part of) the error, and
@@ -303,24 +346,29 @@ class Lexer:
         else:
             i1 = 0
         linestr = "(line %i):" % (l2 + 1)
-        return ("(line %i): %s\n" % (l2 , self.get_line(l2-1))
+        return ("(line %i): %s\n" % (l2, self.get_line(l2 - 1))
               + "%s %s\n" % (linestr, self.get_line(l2))
               + " " * (i1 + len(linestr) + 1) + "^" * (i2 - i1))
+
     def peek(self, forward=0):
-        return self.m_tokens[self.pos+forward]
+        return self.m_tokens[self.pos + forward]
+
     def peek_type(self, forward=0):
-        return self.m_tokens[self.pos+forward][TOKEN_TYPE]
+        return self.m_tokens[self.pos + forward][TOKEN_TYPE]
+
     def peek_string(self, forward=0):
-        return self.m_tokens[self.pos+forward][TOKEN_STRING]
+        return self.m_tokens[self.pos + forward][TOKEN_STRING]
+
     def scan_any(self):
         """scan the next token"""
         self.pos += 1
-        return self.m_tokens[self.pos-1][TOKEN_STRING]
+        return self.m_tokens[self.pos - 1][TOKEN_STRING]
+
     def scan(self, t=None):
         """t is the type of token we expect"""
         if self.m_tokens[self.pos][TOKEN_TYPE] == t:
             self.pos += 1
-            return self.m_tokens[self.pos-1][TOKEN_STRING]
+            return self.m_tokens[self.pos - 1][TOKEN_STRING]
         else:
             # Tested in TestLexer.test_scan
             raise DataparserSyntaxError(self.m_parser(), self.pos,
@@ -328,6 +376,7 @@ class Lexer:
                     'nottoken': t,
                     'foundtoken': self.m_tokens[self.pos][TOKEN_STRING],
                     'type': self.m_tokens[self.pos][TOKEN_TYPE]})
+
     def get_line(self, lineno):
         """line 0 is the first line
         Return an empty string if lineno is out of range.
@@ -343,23 +392,29 @@ class Lexer:
             x += 1
         return self.m_src[idx:x]
 
+
 class Dataparser:
     """
     Parse a lesson file into a parsetree.Program
     """
+
     def __init__(self):
         self.m_filename = None
         self.m_translation_re = re.compile("(?P<varname>\w+)\[(?P<lang>[\w_+]+)\]")
+
     def parse_file(self, filename):
         """We always construct a new parser if we want to parse another
         file. So this method is never called twice for one parser.
         """
         self.m_filename = filename
         self.m_location = os.path.split(filename)[0]
-        self._lexer = Lexer(open(filename, 'rU').read(), self)
+        with open(filename, 'r') as f:
+            self._lexer = Lexer(f.read(), self)
         self.reserved_words = ('_', 'question', 'header')
         self.prog()
+
     def parse_string(self, s, really_filename=False):
+        assert isinstance(s, str)
         if really_filename:
             self.m_filename = really_filename
         else:
@@ -367,6 +422,7 @@ class Dataparser:
         self._lexer = Lexer(s, self)
         self.reserved_words = ('_', 'question', 'header')
         self.prog()
+
     def prog(self):
         """prog: statementlist EOF"""
         self.tree = pt.Program()
@@ -379,10 +435,12 @@ class Dataparser:
             raise DataparserSyntaxError(self, self._lexer.pos,
                     'Expected end of file or statement.')
         self._lexer.scan('EOF')
+
     def statementlist(self):
         """statementlist: (statement+)"""
         while self._lexer.peek_type() == 'NAME':
             self.tree.add_statement(self.statement())
+
     def statement(self, allow_music_shortcut=False):
         """
         statement: assignment | block | include | import
@@ -409,31 +467,34 @@ class Dataparser:
         elif allow_music_shortcut:
             fakt = self.expressionlist()
             if self._lexer.peek_type() == '}':
-                return pt.Assignment(pt.Identifier(u"music"), fakt[0])
+                return pt.Assignment(pt.Identifier("music"), fakt[0])
             else:
                 raise DataparserSyntaxError(self, self._lexer.pos, "The (obsolete) music shortcut construct must be the last statement in a question block.")
         self._lexer.scan_any()
         raise DataparserSyntaxError(self, self._lexer.pos, "Parse error")
+
     def include(self):
-        self._lexer.scan_any() # scan include
-        self._lexer.scan_any() # scan (
+        self._lexer.scan_any()  # scan include
+        self._lexer.scan_any()  # scan (
         try:
             filename = self._lexer.scan('STRING')
         except:
-            print >> sys.stderr, "Warning: The file '%s' uses old style syntax for the include command." % self.m_filename
-            print >> sys.stderr, 'This is not fatal now but will be in the future. You should change the code\nfrom include(filename) to include("filename")\n'
+            print("Warning: The file '%s' uses old style syntax for the include command." % self.m_filename, file=sys.stderr)
+            print('This is not fatal now but will be in the future. You should change the code\nfrom include(filename) to include("filename")\n', file=sys.stderr)
             filename = self._lexer.scan('NAME')
         fn = os.path.join(self.m_location, filename)
         if not os.path.exists(fn):
-            fn = os.path.join(os.getcwdu(), u'exercises/standard/lesson-files', filename)
-        s = open(fn, 'rU').read()
+            fn = os.path.join(os.getcwd(), 'exercises/standard/lesson-files', filename)
+        with open(fn, 'r') as f:
+            s = f.read()
         p = Dataparser()
         p.m_location = self.m_location
         p.parse_string(s)
         self._lexer.scan(')')
         return pt.IncludeStatement(p.tree)
+
     def _import_worker(self, fn1, fn2):
-        self._lexer.scan_any() # scan the 'import' or 'rimport' keyword
+        self._lexer.scan_any()  # scan the 'import' or 'rimport' keyword
         mod_filename = self._lexer.scan_any()
         if (self._lexer.peek_type() == 'NAME'
             and self._lexer.peek_string() == 'as'):
@@ -449,22 +510,25 @@ class Dataparser:
         else:
             p.parse_file(fn2)
         return pt.Assignment(pt.Identifier(mod_name), p.tree)
+
     def do_import(self):
         return self._import_worker(
-            os.path.join(os.getcwdu(), "exercises", "standard", "lib"),
+            os.path.join(os.getcwd(), "exercises", "standard", "lib"),
             os.path.join(self.m_location, "..", "lib"))
+
     def do_rimport(self):
         return self._import_worker(
             os.path.join(self.m_location, "..", "lib"),
-            os.path.join(os.getcwdu(), "exercises", "standard", "lib"))
+            os.path.join(os.getcwd(), "exercises", "standard", "lib"))
+
     def assignment(self):
         """NAME "=" expression ("," expression)* """
         npos = self._lexer.pos
-        name = self._lexer.scan_any()#('NAME')
+        name = self._lexer.scan_any()  # ('NAME')
         if name in self.reserved_words:
             # do "question = 1" to raise this exception.
             raise AssignmentToReservedWordException(self, npos, name)
-        self._lexer.scan_any()#('=')
+        self._lexer.scan_any()  # ('=')
         fpos = self._lexer.pos
         expressionlist = self.expressionlist()
         m = self.m_translation_re.match(name)
@@ -477,6 +541,7 @@ class Dataparser:
             return pt.Assignment(pt.Identifier(name), expressionlist[0])
         else:
             return pt.Assignment(pt.Identifier(name), expressionlist)
+
     def expression(self):
         """expression: atom
               ("+" atom
@@ -507,6 +572,7 @@ class Dataparser:
                 break
             peek = self._lexer.peek_type()
         return expression
+
     def expressionlist(self):
         """expressionlist: expression ("," expression)* """
         # use tmp var just in case an atom in the future is more than
@@ -518,6 +584,7 @@ class Dataparser:
             self._lexer.scan_any()
             expressionlist.append(self.expression())
         return expressionlist
+
     def atom(self):
         """atom: INTEGER | FLOAT | STRING | NAME | FUNCTIONCALL"""
         npos = self._lexer.pos
@@ -539,13 +606,14 @@ class Dataparser:
                 name += "." + self._lexer.scan('NAME')
             return pt.Identifier(name)
         else:
-            #print "FIXME: have no idea how to raise this exception"
+            # print "FIXME: have no idea how to raise this exception"
             raise DataparserSyntaxError(self, npos + 1,
                 "Expected STRING, INTEGER or NAME+'('")
+
     def functioncall(self):
         """functioncall: NAME "(" expressionlist ")" """
         npos = self._lexer.pos
-        name = self._lexer.scan_any()#'NAME')
+        name = self._lexer.scan_any()  # 'NAME')
         self._lexer.scan('(')
         if self._lexer.peek_type() == ')':
             # functioncall()
@@ -558,10 +626,11 @@ class Dataparser:
             ret = pt.FunctionCall(self, name, arglist)
         ret.m_tokenpos = npos
         return ret
+
     def block(self):
         """block: NAME "{" assignments"}" """
         block = pt.Block(self._lexer.scan_any())
-        self._lexer.scan_any() # scan '{'
+        self._lexer.scan_any()  # scan '{'
         # FIXME this is not nice, but we need it since we have
         # allowed question blocks with the shortcut where they
         # omit "music ="
@@ -569,6 +638,7 @@ class Dataparser:
             block.add_statement(self.statement(allow_music_shortcut=True))
         self._lexer.scan("}")
         return block
+
     def named_block(self):
         blocktype = self._lexer.scan('NAME')
         name = self._lexer.scan('NAME')
@@ -578,4 +648,3 @@ class Dataparser:
             block.add_statement(self.statement())
         self._lexer.scan("}")
         return block
-
