@@ -1,6 +1,6 @@
 # GNU Solfege - free ear training software
 # vim: set fileencoding=utf-8 :
-# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2011 Tom Cato Amundsen
+# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2011, 2016 Tom Cato Amundsen
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -16,12 +16,9 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 
-from __future__ import absolute_import
-
 import hashlib
 import logging
 import os
-import pickle
 import shutil
 import sqlite3
 import time
@@ -43,21 +40,26 @@ def hash_lessonfile_text(s):
     lines = s.split("\n")
     lines = [x for x in lines if (not x.startswith("#") and bool(x))]
     sha1 = hashlib.sha1()
-    sha1.update("\n".join(lines))
+    sha1.update("\n".join(lines).encode("utf-8"))
     return sha1.hexdigest()
 
 
 def hash_of_lessonfile(filename):
-    assert isinstance(filename, unicode)
-    return hash_lessonfile_text(open(lessonfile.uri_expand(filename), 'r').read())
+    assert isinstance(filename, str)
+    with open(lessonfile.uri_expand(filename), 'r') as f:
+        return hash_lessonfile_text(f.read())
 
 
 class DB(object):
-    type_int_dict = {int: 0, unicode: 1, float: 2}
-    int_type_dict = {0: int, 1: unicode, 2: float}
+    type_int_dict = {int: 0, str: 1, float: 2}
+    int_type_dict = {0: int, 1: str, 2: float}
+
     class VariableTypeError(Exception): pass
+
     class VariableUndefinedError(Exception): pass
+
     class FileNotInDB(Exception): pass
+
     def __init__(self, callback=None, profile=None):
         """
         callback is called to display progress when scanning lesson files.
@@ -73,7 +75,7 @@ class DB(object):
             statistics_filename = self.get_statistics_filename()
             # We make a backup of the statistics file since Solfege 3.21.2 will
             # delete the statistics for the elembuilder exercise.
-            bk_filename = u"%s.pre-3.21.2.backup" % statistics_filename
+            bk_filename = "%s.pre-3.21.2.backup" % statistics_filename
             if os.path.exists(statistics_filename) and not os.path.exists(bk_filename):
                 shutil.copyfile(statistics_filename, bk_filename)
             head, tail = os.path.split(statistics_filename)
@@ -97,8 +99,10 @@ class DB(object):
             self.upgrade_to_version_2()
         if db_ver < 3:
             self.upgrade_to_version_3()
-        self.set_variable("database_version", 3)
+        self.upgrade_to_version_4()
+        self.set_variable("database_version", 4)
         self.conn.commit()
+
     def insert_file(self, filename):
         assert lessonfile.is_uri(filename) or os.path.isabs(filename)
         self.conn.execute("insert into lessonfiles "
@@ -106,19 +110,23 @@ class DB(object):
                 "values (?, ?)",
                 (filename, hash_of_lessonfile(filename)))
         return self.get_fileid(filename)
+
     @staticmethod
     def get_noprofile_statistics_filename():
         return os.path.join(filesystem.app_data(), "statistics.sqlite")
+
     def get_statistics_filename(self):
         if self.m_profile:
             return os.path.join(filesystem.app_data(), "profiles", self.m_profile, "statistics.sqlite")
         else:
             return self.get_noprofile_statistics_filename()
+
     def reset_database(self):
         self.conn.close()
         os.remove(self.get_statistics_filename())
         self.conn = sqlite3.connect(self.get_statistics_filename())
         self.create_tables()
+
     def remove_tables(self):
         """
         Remove the tables if they exist.
@@ -130,6 +138,7 @@ class DB(object):
         # Current Solfege does not use this table, but we did until
         # the 3.17.0 release. So we remove it if it exists.
         self.conn.execute("drop table if exists tests")
+
     def create_tables(self):
         self.conn.execute(
             "create table if not exists lessonfiles ( "
@@ -152,6 +161,7 @@ class DB(object):
             "(fileid int, timestamp int, sessiontype int, "
             "unique (fileid, timestamp))")
         self.set_variable("database_version", 3)
+
     def drop_if_has_uuid_column(self):
         """
         Drop the database and return True if the lessonfiles table
@@ -162,17 +172,19 @@ class DB(object):
         columns = [x[1] for x in
             self.conn.execute('pragma table_info(lessonfiles)').fetchall()]
         # true if the database is created by solfege 3.15.0 - 3.15.2
-        if u'uuid' in columns:
+        if 'uuid' in columns:
             logging.debug("statistics: dropping tables because the database is from solfege-3.15.0-3.15.2")
             self.remove_tables()
             self.create_tables()
             return True
+
     def _get_table_names(self):
         """
         Return a set of all table names in the database.
         """
         tables = self.conn.execute('select name from sqlite_master where type="table";').fetchall()
         return set([x[0] for x in tables])
+
     def sanity_check(self):
         """
         The content of the tables 'sessions', 'sessioninfo' and 'lessonfiles'
@@ -189,7 +201,7 @@ class DB(object):
         for t in ('lessonfiles', 'sessions', 'sessioninfo', 'variables'):
             if t not in tables:
                 missing_table = True
-                break    
+                break
         else:
             missing_table = False
         # This is a sanity check. If one of the tables are missing, then
@@ -198,6 +210,7 @@ class DB(object):
             self.remove_tables()
             self.create_tables()
             return True
+
     def get_fileid(self, filename):
         assert lessonfile.is_uri(filename) or os.path.isabs(filename), filename
         row = self.conn.execute(
@@ -208,6 +221,7 @@ class DB(object):
             return row[0]
         else:
             raise self.FileNotInDB(filename)
+
     def get_session_count(self, fileid):
         """
         Return an int telling the number of practise sessions we have
@@ -215,6 +229,7 @@ class DB(object):
         """
         count = self.conn.execute("select count(distinct timestamp) from sessions where fileid=?", (fileid,)).fetchone()[0]
         return count if count is not None else 0
+
     def get_first_timestamp(self, fileid):
         """
         Return the timestamp (as a python datetime object) of the first
@@ -223,6 +238,7 @@ class DB(object):
         """
         return self.conn.execute("select min(timestamp) from sessions where fileid=?",
                           (fileid,)).fetchone()[0]
+
     def get_last_timestamp(self, fileid):
         """
         Return the timestamp (as a python datetime object) of the last
@@ -231,12 +247,28 @@ class DB(object):
         """
         return self.conn.execute("select max(timestamp) from sessions where fileid=?",
                           (fileid,)).fetchone()[0]
+
     def delete_statistics(self, filename):
         fileid = self.get_fileid(filename)
         cursor = self.conn.cursor()
         cursor.execute("delete from sessions where fileid=?", (fileid,))
         cursor.execute("delete from sessioninfo where fileid=?", (fileid,))
+        cursor.execute("delete from toneincontext where fileid=?", (fileid,))
         self.conn.commit()
+
+    def upgrade_to_version_4(self):
+        try:
+            if self.get_variable("database_version") >= 4:
+                return
+        except self.VariableUndefinedError:
+            pass
+        self.conn.execute(
+            "create table if not exists toneincontext ( "
+            " fileid int, "
+            " timestamp int, "
+            " answerkey int, "
+            " guessedkey int "
+            ")")
     def upgrade_to_version_3(self):
         try:
             if self.get_variable("database_version") >= 3:
@@ -249,6 +281,7 @@ class DB(object):
                     self.delete_statistics(fn)
                 except self.FileNotInDB:
                     pass
+
     def upgrade_to_version_2(self):
         """
         After releasing 3.16.0 I found that having separate tests and
@@ -275,8 +308,8 @@ class DB(object):
                     "(fileid, timestamp, sessiontype) "
                     "values (?, ?, ?)",
                     (fileid, timestamp, 1))
-            except sqlite3.IntegrityError, e:
-                logging.error(u"%s, %s, %s", e, fileid, timestamp)
+            except sqlite3.IntegrityError as e:
+                logging.error("%s, %s, %s", e, fileid, timestamp)
         # Insert all practise sessions into sessioninfo
         for fileid, timestamp in self.conn.execute("select distinct fileid, timestamp from sessions"):
             # This can fail because 3.14.11 and earlier had a bug where the same
@@ -286,8 +319,8 @@ class DB(object):
                     "(fileid, timestamp, sessiontype) "
                     "values (?, ?, ?)",
                     (fileid, timestamp, 0))
-            except sqlite3.IntegrityError, e:
-                logging.error(u"%s, %s, %s", e, fileid, timestamp)
+            except sqlite3.IntegrityError as e:
+                logging.error("%s, %s, %s", e, fileid, timestamp)
         # Copy all tests from the obsolete "tests" table to "sessions"
         for a, b, c, d in self.conn.execute("select fileid, timestamp, answerkey, guessed from tests"):
             # This can fail because 3.14.11 and earlier had a bug where the same
@@ -297,8 +330,8 @@ class DB(object):
                     "(fileid, timestamp, answerkey, guessed) "
                     "values (?, ?, ?, ?)",
                     (a, b, c, d))
-            except sqlite3.IntegrityError, e:
-                logging.error(u"%s, %s, %s", e, fileid, timestamp)
+            except sqlite3.IntegrityError as e:
+                logging.error("%s, %s, %s", e, fileid, timestamp)
         for fileid, filename in self.conn.execute(
                 "select sessioninfo.fileid, lessonfiles.filename "
                 "from sessioninfo, lessonfiles "
@@ -318,6 +351,7 @@ class DB(object):
                     p.parse_file(filename)
                     self.cache_new_test_result(filename, timestamp,
                         p.get_test_requirement(), p.get_test_num_questions())
+
     def cache_new_test_result(self, filename, timestamp,
                               required, num_questions):
         """
@@ -348,6 +382,7 @@ class DB(object):
              test_result >= required,
              fileid,))
         self.conn.commit()
+
     def get_test_status(self, filename):
         """
         Return a tuple saying if the test was passed or not, and the result:
@@ -363,6 +398,7 @@ class DB(object):
             return row
         except self.FileNotInDB:
             return None, None
+
     def validate_stored_statistics(self, filename):
         """
         Insert the filename and hash of the file into the lessonfiles
@@ -385,20 +421,28 @@ class DB(object):
         if not row:
             # Ususally the filename exists in the database, but when running
             # the test suite, it does not, so we have the code here to add it.
+            logging.debug("file %s not in database. Adding", filename)
             self.insert_file(filename)
             self.conn.commit()
         else:
             hashvalue, fileid = row
             if hashvalue != cur_lessonfile_hash_value:
+                logging.debug(" hashvalue != cur_lessonfile_hash_value")
                 replaces = solfege.lessonfile.infocache.get(filename, 'replaces')
                 if not isinstance(replaces, list):
                     replaces = [replaces]
                 if not hashvalue in replaces:
+                    logging.debug("  hash value changed. Deleing statistics")
                     solfege.db.delete_statistics(filename)
                     cursor.execute("update lessonfiles "
                         "set hash=?, test_passed=?, test_result=0.0 where fileid=?",
                         (cur_lessonfile_hash_value, None, fileid))
                     self.conn.commit()
+                else:
+                    logging.debug("  hash value matched")
+            else:
+                logging.debug(" hashvalue matched")
+
     def get_statistics_info(self):
         """
         Return information about the data installed.
@@ -416,6 +460,7 @@ class DB(object):
         return {'exercises': different_ex,
                 'practise_count': session_count,
                 'test_count': test_count}
+
     def _recent(self, count, sessiontype):
         """
         sessiontype 0 == normal statistics
@@ -433,10 +478,13 @@ class DB(object):
             if len(filenames) == count:
                 break
         return filenames
+
     def recent(self, count):
         return self._recent(count, 0)
+
     def recent_tests(self, count):
         return self._recent(count, 1)
+
     def set_variable(self, name, value):
         """
         raise DB.VariableTypeError if the type we set is different from
@@ -455,11 +503,12 @@ class DB(object):
             self.conn.execute("insert into variables "
                     "(variable_name, type, value) "
                     "values (?, ?, ?)",
-                    (name, self.type_int_dict[type(value)], unicode(value)))
+                    (name, self.type_int_dict[type(value)], str(value)))
         else:
             self.conn.execute("update variables "
                 "set value=? where variable_name=?",
-                (unicode(value), name))
+                (str(value), name))
+
     def get_variable(self, name):
         cursor = self.conn.execute('select type, value from variables where variable_name=?', (name,))
         try:
@@ -467,6 +516,7 @@ class DB(object):
         except TypeError:
             raise self.VariableUndefinedError()
         return self.int_type_dict[type](value)
+
     def del_variable(self, name):
         """
         Delete the variable, raise DB.VariableUndefinedError if the variable
@@ -480,10 +530,12 @@ class DB(object):
 
 
 class AbstractStatistics(object):
+
     def __init__(self, teacher):
         self.m_t = teacher
         self.m_timestamp = None
         self.m_test_mode = False
+
     def int_if_int(self, s):
         """
         Return an int if the can be converted to an int.
@@ -493,6 +545,7 @@ class AbstractStatistics(object):
             return int(s)
         except ValueError:
             return s
+
     def get_keys(self, all_keys=False):
         """
         Return the keys for all questions that have been answered correctly.
@@ -515,7 +568,8 @@ class AbstractStatistics(object):
             c = [x[0] for x in list(c)]
         v = [self.int_if_int(x) for x in c]
         v.sort()
-        return [unicode(x) for x in v]
+        return [str(x) for x in v]
+
     def get_statistics(self, seconds):
         """
         return a dict with statistics more recent than 'seconds' seconds.
@@ -542,11 +596,13 @@ class AbstractStatistics(object):
             ret.setdefault(answer, {})
             ret[answer][guess] = count
         return ret
+
     def reset_session(self):
         """
         Start a new practise session.
         """
         self.m_timestamp = int(time.time())
+
     def reset_custom_mode_session(self, filename):
         """
         Exercises that show statistics in custom_mode should call this
@@ -556,8 +612,14 @@ class AbstractStatistics(object):
         fileid = solfege.db.get_fileid(filename)
         solfege.db.conn.execute("delete from sessions "
                           "where fileid=? and timestamp=1", (fileid,))
+        solfege.db.conn.execute(
+            "delete from toneincontext "
+            "where fileid=?", (fileid,))
+        print("Deletet statistics for", filename)
+
     def enter_test_mode(self):
         self.m_test_mode = True
+
     def exit_test_mode(self):
         """
         If the user cancels a test, the answers give will be recorded.
@@ -570,6 +632,7 @@ class AbstractStatistics(object):
                 self.m_t.m_P.get_test_requirement(),
                 self.m_t.m_P.get_test_num_questions())
         self.reset_session()
+
     def _add(self, question, answer):
         """
         Register that for the question 'question' the user answered 'answer'.
@@ -597,31 +660,35 @@ class AbstractStatistics(object):
         row = cursor.execute(
                 "select count from sessions where fileid=? and timestamp=? "
                 "and answerkey=? and guessed=?",
-                (fileid, self.m_timestamp, unicode(question), unicode(answer))).fetchone()
+                (fileid, self.m_timestamp, str(question), str(answer))).fetchone()
         if not row:
             cursor.execute(
                 "insert into sessions "
                 "(fileid, timestamp, answerkey, guessed, count) "
                 "values(?, ?, ?, ?, ?)",
                 (fileid, self.m_timestamp,
-                 unicode(question), unicode(answer), 1))
+                 str(question), str(answer), 1))
         else:
             assert cursor.fetchone() is None
             cursor.execute(
                 "update sessions set count=? where "
                 "fileid=? and timestamp=? and answerkey=? and guessed=?",
                 (row[0] + 1, fileid,
-                 self.m_timestamp, unicode(question), unicode(answer)))
+                 self.m_timestamp, str(question), str(answer)))
         solfege.db.conn.commit()
+
     def add_wrong(self, question, answer):
         self._add(question, answer)
+
     def add_correct(self, answer):
         self._add(answer, answer)
+
     def get_last_test_result(self):
         """
         Return the test result of the last test ran for this lesson file.
         """
         return solfege.db.last_test_result(self.m_t.m_P.m_filename)
+
     def get_percentage_correct(self):
         """Will return a 0 <= value <= 100.0 that say how many percent is
         correct in this session.
@@ -635,6 +702,7 @@ class AbstractStatistics(object):
             num_asked = 0
             return 0
         return 100.0 * num_correct / num_asked
+
     def get_num_correct_for_key(self, seconds, key):
         """
         Return the number of correct answers for the given key 'key' the
@@ -653,6 +721,7 @@ class AbstractStatistics(object):
         if ret:
             return ret
         return 0
+
     def get_num_guess_for_key(self, seconds, key):
         """
         See get_num_correct_for_key docstring.
@@ -667,6 +736,7 @@ class AbstractStatistics(object):
         if ret:
             return ret
         return 0
+
     def get_num_guess(self, seconds):
         fileid = solfege.db.get_fileid(self.m_t.m_P.m_filename)
         if seconds == -1:
@@ -678,6 +748,7 @@ class AbstractStatistics(object):
         if ret:
             return ret
         return 0
+
     def get_num_correct(self, seconds):
         fileid = solfege.db.get_fileid(self.m_t.m_P.m_filename)
         if seconds == -1:
@@ -689,6 +760,7 @@ class AbstractStatistics(object):
         if ret:
             return ret
         return 0
+
     def iter_test_results(self):
         """
         Iterate test results for the associated lesson file, newest
@@ -713,6 +785,7 @@ class AbstractStatistics(object):
 
 
 class LessonStatistics(AbstractStatistics):
+
     def key_to_pretty_name(self, key):
         def ff(x):
             try:
@@ -727,7 +800,9 @@ class LessonStatistics(AbstractStatistics):
                 return ff(question.name)
         return ff(key)
 
+
 class IntervalStatistics(AbstractStatistics):
+
     def get_keys(self, all_keys=False):
         # FIXME we have to check that all keys are integers, and filter
         # out those that are not. I don't know how, but I ended up with
@@ -742,15 +817,18 @@ class IntervalStatistics(AbstractStatistics):
                 return False
         v = AbstractStatistics.get_keys(self, all_keys)
         return [x for x in v if isinteger(x)]
+
     def key_to_pretty_name(self, key):
         return utils.int_to_intervalname(int(key), 1, 1)
 
+
 class HarmonicIntervalStatistics(IntervalStatistics):
+
     def key_to_pretty_name(self, key):
         return utils.int_to_intervalname(int(key), 1, 0)
 
+
 class IdToneStatistics(LessonStatistics):
+
     def key_to_pretty_name(self, key):
         return mpd.MusicalPitch.new_from_notename(key).get_user_notename()
-
-
